/*
 *  YASS 1.07
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include "util.h"
#include "global_var.h"
#include "kword.h"
#include "tuple.h"
#include "assemble.h"
#include "prdyn.h"
#include "proba.h"
#include "display.h"
#include "regroup.h"

/******************************
 * MACRO PARSING COMMAND LINE *
 ******************************/


#define MAXHW 24

void USAGE() {     
  SETBCOLOR(stderr,GREEN);                                                            
  fprintf(stderr,"*  USAGE:\n"); 
  RESET(stderr);
  fprintf(stderr,"  %s [options] { fasta.fas | fasta.fas multifasta.fas }\n",PACKAGE);   
  fprintf(stderr,"      -h       display this Help screen\n");
  fprintf(stderr,"      -d <int>    0 : Display alignment positions (default)\n"
	         "                  1 : Display alignment positions + alignments + statistics\n"
	         "                  2 : Display blast-like tabular output\n"
	         "                  3 : Display light tabular output (better for post-processing)\n");

  fprintf(stderr,"      -r <int>    0 : process forward (query) strain\n" 
	         "                  1 : process Reverse complement strain\n"
	         "                  2 : process both forward and Reverse complement strains\n");
  fprintf(stderr,"      -o <str> Output file\n");
  fprintf(stderr,"      -v       display the current Version\n");
  fprintf(stderr,"                                                                               \n");
  fprintf(stderr,"      -M <int> select a scoring Matrix (0 by default):\n"
	         "               [Match,Transition,Transversion],(Gopen,Gext)\n"
                 "               0 : [1,-2,-3],( -8,-2)   1 : [3,-1,-2],(-12,-4)\n"
                 "               2 : [5,-3,-4],(-16,-4)   3 : [3,-1,-4],(-16,-4)\n");
  fprintf(stderr,"      -C <int>[,<int>[,<int>[,<int>]]] reset match,mismatch,transition,other\n"
	         "                                       Costs (penalties)\n");
  fprintf(stderr,"      -G <int>,<int> reset Gap opening/extension penalties\n");  
  fprintf(stderr,"      -L <dbl>,<dbl> reset Lambda and K parameters of Gumbel law\n");
  fprintf(stderr,"      -X <int> Xdrop threshold score (default %i)\n",gp_xdrop);
  fprintf(stderr,"      -E <int> E-value threshold (default %1.2g)\n",gp_expectation_value);
  fprintf(stderr,"                                                                               \n");
  fprintf(stderr,"      -O <int> limit number of Output alignments (default %i)\n",gp_nbmaxlines); 
  fprintf(stderr,"      -S <int> Select query sequence from multi-fasta query file (default %i)\n", gp_selection_fasta);
  fprintf(stderr,"      -T <int> forbid aligning too close regions (e.g. Tandem repeats)\n"
	         "               valid for single sequence comparison only (default %i bp)\n", gp_distdiag);  
  fprintf(stderr,"                                                                               \n");
  fprintf(stderr,"      -s <int> group Size              (default %i bp)\n",gp_minimum_score_seeds);
  fprintf(stderr,"      -p <str> seed Pattern            (default %s)\n"
	         "                 * use \'#\' for match \n"
                 "                 * use \'@\' for match or transition \n"
                 "                 - example \"%s file.fas -p %s \n",gp_motif,PACKAGE,gp_motif);
  fprintf(stderr,"                                                                               \n");
  fprintf(stderr,"      -e <dbl> minimal allowed Entropy of trinucleotide distribution  \n"
	         "               (ranging between 0 and 6, default %1.2f)               \n",gp_entropy_min);
  fprintf(stderr,"      -a <int> statistical tolerance Alpha (%%) (default %i%%)\n",gp_alpha_percent);
  fprintf(stderr,"      -i <int> Indel rate (%%)                  (default %i%%)\n",gp_indels_percent);  
  fprintf(stderr,"      -m <int> Mutation rate (%%)               (default %i%%)\n",gp_mutations_percent);
  fprintf(stderr,"                                                                               \n");
  fprintf(stderr,"      -W <int>,<int> Window <min,max> range for post-processing and grouping\n"
	         "                     alignments (default <%i,%i>) \n",gp_win_min,gp_win_max);
  fprintf(stderr,"      -w <int> Window size coefficient for post-processing and grouping        \n"
                 "               alignments (default %i). NOTE : -w 0 disables post-processing\n",gp_win_mul);
  RESET(stderr);
  RESET(stdout);
  exit(-1);
}

void VERS() {
  fprintf(stderr,"version %s\n",VERSION);
  exit(-1);
}



void ERROR(char * E1,char * E2,char * E3)                    
{     
  SETBCOLOR(stderr,RED);
  fprintf(stderr,"** ERROR:\n  ");   
  RESET(stderr);
  fprintf(stderr,E1);
  fprintf(stderr,E2);
  fprintf(stderr,E3);
  fprintf(stderr,"\n");
  USAGE(); 
}                                         

#define PARSEFLAG(VAR,VALUE) {          \
       (VAR) = (VALUE);                 \
       goto endofparam;                 \
}

#define PARSEPATTERN()  {                                                            \
     i++;                                                                            \
     if (i >= argc)                                                                  \
        ERROR("\"",argv[i-1],"\" found without motif");                              \
     if (!(strchr(argv[i],'#')||strchr(argv[i],'@')))                                \
        ERROR(" pattern \"",argv[i],"\"  doesn't have any \"#\" or \"@\" character");\
     strncpy(gp_motif,argv[i],255);                                                  \
     ComputeLength();                                                                \
     if ( gp_seed_bitweight   > MAXHW)                                               \
        ERROR(" pattern \"",argv[i-1],"\" has too many \"#\" and \"@\" characters"); \
     overrides_weight = 1 ;                                                          \
     goto endofparam;                                                                \
}

#define PARSEOUTFILE() {                                                       \
     i++;                                                                      \
     if (i >= argc)                                                            \
        ERROR("\"",argv[i-1],"\" found without filename");                     \
     gv_outstream = fopen(argv[i],"w");                                        \
     if (!gv_outstream)                                                        \
	ERROR("Can't create file \"",argv[i],"\"");                            \
     goto endofparam;                                                          \
}


#define PARSEINT(VAR,MN,MX) {                                                         \
     i++;                                                                             \
     if (i>=argc)                                                                     \
        ERROR("\"",argv[i-1],"\" found without argument");                            \
     i_tmp = strtol(argv[i],NULL,10);                                                 \
     if (errno != 0 )                                                                 \
        ERROR("\"",argv[i-1],"\" is not followed by an integer");                     \
     if (i_tmp>MX || i_tmp <MN)                                                       \
        ERROR("\"",argv[i-1],"\" is followed by an integer outside the valid range"); \
     VAR = i_tmp;                                                                     \
     goto endofparam;                                                                 \
}


#define PARSEDINT(VAR1,MIN1,MAX1,VAR2,MIN2,MAX2) {                                    \
     i++;                                                                             \
     if (i>=argc)                                                                     \
        ERROR("\"",argv[i-1],"\" found without argument");                            \
     i_tmp = sscanf(argv[i],"%i,%i",&VAR1,&VAR2);                                     \
     if (i_tmp != 2 )                                                                 \
        ERROR("\"",argv[i-1],"\" is not followed by an <int>,<int>");                 \
     if (VAR1 < MIN1 || VAR1 > MAX1 || VAR2 < MIN2 || VAR2 > MAX2)                    \
        ERROR("\"",argv[i-1],"\" is followed by integers out of the range");          \
     goto endofparam;                                                                 \
}


#define PARSEDOUBLE(VAR,MN,MX) {                                                      \
     i++;                                                                             \
     if (i>=argc)                                                                     \
        ERROR("\"",argv[i-1],"\" found without argument");                            \
     d_tmp = strtod(argv[i],NULL);                                                    \
     if (errno != 0 )                                                                 \
        ERROR("\"",argv[i-1],"\" is not followed by a double") ;                      \
     if (d_tmp>MX || d_tmp <MN)                                                       \
        ERROR("\"",argv[i-1],"\" is followed by a double outside the valid range");   \
     VAR = d_tmp;                                                                     \
     goto endofparam;                                                                 \
	}

#define PARSEDDOUBLE(VAR1,MIN1,MAX1,VAR2,MIN2,MAX2) {                                 \
     i++;                                                                             \
     if (i>=argc)                                                                     \
        ERROR("\"",argv[i-1],"\" found without argument");                            \
     i_tmp = sscanf(argv[i],"%lf,%lf",&VAR1,&VAR2);                                   \
     if (i_tmp != 2 )                                                                 \
        ERROR("\"",argv[i-1],"\" is not followed by an <dbl>,<dbl>");                 \
     if (VAR1 < MIN1 || VAR1 > MAX1 || VAR2 < MIN2 || VAR2 > MAX2)                    \
        ERROR("\"",argv[i-1],"\" is followed by integers out of the range");          \
     goto endofparam;                                                                 \
}
#define PARSECOST() {                                         \
 i++;                                                         \
 if (i>=argc)                                                 \
        ERROR("\"",argv[i-1],"\" found without argument");    \
 if ((s_tmp=strchr(argv[i],','))!=NULL) {                     \
   *s_tmp=0;                                                  \
   gp_cost_match = atoi (argv[i]);                            \
   argv[i]=s_tmp+1;                                           \
   if ((s_tmp=strchr(argv[i],','))!=NULL) {                   \
     *s_tmp=0;                                                \
     gp_cost_mismatch = atoi (argv[i]);                       \
     argv[i]=s_tmp+1;                                         \
     if ((s_tmp=strchr(argv[i],','))!=NULL) {                 \
      *s_tmp=0;                                               \
      gp_cost_transition = atoi (argv[i]);                    \
      argv[i]=s_tmp+1;                                        \
      gp_cost_other =  atoi (argv[i]);                        \
     }                                                        \
     else                                                     \
       gp_cost_transition = atoi (argv[i]);                   \
   }                                                          \
   else                                                       \
     gp_cost_mismatch = atoi (argv[i]);                       \
 }                                                            \
 else                                                         \
   gp_cost_match = atoi (argv[i]);                            \
 if ( gp_cost_transition>0 ||gp_cost_mismatch>0||gp_cost_other >0)         \
 fprintf(stderr,"warning %s converts transition/gp_cost_mismatch/"         \
	 "gp_cost_other to negative value\n",argv[0]);                     \
 if ( gp_cost_match<0){fprintf(stderr,"warning %s converts gp_cost_match"  \
			       " to potive value",argv[0]);                \
                       gp_cost_match=ABS(gp_cost_match);                   \
 }                                                                         \
 if ( gp_cost_transition>0) gp_cost_transition=-ABS(gp_cost_transition);   \
 if ( gp_cost_mismatch>0)   gp_cost_mismatch=-ABS(gp_cost_mismatch);       \
 if ( gp_cost_other >0)     gp_cost_other=-ABS(gp_cost_other);             \
 if ( gp_cost_match<-1000  || gp_cost_match>1000                           \
      ||gp_cost_transition<-1000                                           \
      ||gp_cost_mismatch<-1000                                             \
      ||gp_cost_other<-1000  || gp_cost_other>1000 )                       \
     ERROR("\"",argv[i-1],"\" is followed by integers out of the range");  \
     goto endofparam;                                                      \
}

#define PARSEPARAM() {                                                    \
	if(!strcmp(argv[i],"-a")||!strcmp(argv[i],"--alpha"))             \
		PARSEINT(gp_alpha_percent,1,99);                          \
        if(!strcmp(argv[i],"-i")||!strcmp(argv[i],"--indels"))            \
		PARSEINT(gp_indels_percent,0,100);                        \
        if(!strcmp(argv[i],"-m")||!strcmp(argv[i],"--mutations"))         \
		PARSEINT(gp_mutations_percent,0,99);                      \
        if(!strcmp(argv[i],"-s")||!strcmp(argv[i],"--minimumscoreseeds")) \
		PARSEINT(gp_minimum_score_seeds,0,0x7fffffff);            \
        if(!strcmp(argv[i],"-M")||!strcmp(argv[i],"--Matrix"))            \
		PARSEINT(matrix,0,3);                                     \
        if(!strcmp(argv[i],"-T")||!strcmp(argv[i],"--Tandems"))           \
		PARSEINT(gp_distdiag,0,0x7fffffff);                       \
        if(!strcmp(argv[i],"-X")||!strcmp(argv[i],"--Xdrop"))             \
		PARSEINT(gp_xdrop,0,100);                                 \
        if(!strcmp(argv[i],"-O")||!strcmp(argv[i],"--nbO"))               \
		PARSEINT(gp_nbmaxlines,1,0x7fffffff);                     \
        if(!strcmp(argv[i],"-S")||!strcmp(argv[i],"--nbO"))               \
		PARSEINT(gp_selection_fasta,1,0x7fffffff);                \
        if(!strcmp(argv[i],"-G")||!strcmp(argv[i],"--Gaps"))              \
	        PARSEDINT(gp_cost_gap_opened,1,0x7fffffff,                \
			  gp_cost_gap_continued,1,0x7fffffff);            \
        if(!strcmp(argv[i],"-C")||!strcmp(argv[i],"--Costs"))             \
	        PARSECOST();                                              \
	if(!strcmp(argv[i],"-W")||!strcmp(argv[i],"--Windows"))           \
	        PARSEDINT(gp_win_min,0,0x7fffffff,                        \
			  gp_win_max,0,0x7fffffff);                       \
	if(!strcmp(argv[i],"-w")||!strcmp(argv[i],"--window"))            \
	        PARSEINT(gp_win_mul,0,0x7fffffff);                        \
        if(!strcmp(argv[i],"-E")||!strcmp(argv[i],"--Expectation"))       \
	        PARSEDOUBLE(gp_expectation_value,1e-300,1e+10);           \
        if(!strcmp(argv[i],"-L")||!strcmp(argv[i],"--Lambda"))            \
	        PARSEDDOUBLE(gp_lambda_blast,1e-9,1e+9,                   \
			     gp_k_blast,     1e-9,1e+9);                  \
        if(!strcmp(argv[i],"-r")||!strcmp(argv[i],"--reverse"))           \
		PARSEINT(gp_reverse,0,2);                                 \
        if(!strcmp(argv[i],"-d")||!strcmp(argv[i],"--display"))           \
		PARSEINT(gp_display,0,3);                                 \
	if(!strcmp(argv[i],"-p")||!strcmp(argv[i],"--pattern"))           \
		PARSEPATTERN();                                           \
	if(!strcmp(argv[i],"-e")||!strcmp(argv[i],"--entropy"))           \
		PARSEDOUBLE(gp_entropy_min,0.0,6.0);                      \
        if(!strcmp(argv[i],"-o")||!strcmp(argv[i],"--output"))            \
		PARSEOUTFILE();                                           \
	if(!strcmp(argv[i],"-h")||!strcmp(argv[i],"--help"))              \
		USAGE();                                                  \
	if(!strcmp(argv[i],"-v")||!strcmp(argv[i],"--version"))           \
		VERS();                                                   \
        ERROR("\"",argv[i],"\" is not a valid argument");                 \
}


#define PARSENAME() {                                                     \
	if (gp_nbfiles == 2)                                              \
	  ERROR("Too many files : \"",argv[i],"\" is the third one ..."); \
        strncpy(gp_files[gp_nbfiles],argv[i],255);                        \
        if (!(f=fopen(gp_files[gp_nbfiles],"r")))                         \
     	  ERROR("Can't open file \"",argv[i],"\"");                       \
        fclose(f);                                                        \
	gp_nbfiles++;                                                     \
	goto endofparam;                                                  \
} 


#define SCANARG()                                              \
 for(i=1;i<argc;i++){                                          \
    if(argv[i][0] == '-' )                                     \
        PARSEPARAM()                                           \
    else                                                       \
        PARSENAME()                                            \
    endofparam: ;                                              \
 }                                                             \
                                                               \
ComputeLength();                                               \
/* files */                                                    \
if (gp_nbfiles == 0)                                           \
   ERROR("No file found"," ... ","give at least one file ...");\
if ( gp_seed_bitweight >= 2*gp_minimum_score_seeds )           \
   ERROR("weight needs to be under s","",".");                 \
/* gap costs */                                                \
if(gp_cost_gap_opened == 0 && gp_cost_gap_continued == 0){     \
   gp_cost_gap_opened     = INDELSTABLE[matrix][0];            \
   gp_cost_gap_continued  = INDELSTABLE[matrix][1];            \
}else{                                                         \
   gp_cost_gap_opened     = ABS(gp_cost_gap_opened);           \
   gp_cost_gap_continued  = ABS(gp_cost_gap_continued);        \
}                                                              \
/* subs matrix */                                              \
for (x=0;x<16;x++)                                             \
 for (y=0;y<16;y++){                                           \
  if (x<6 && y<6)                                              \
   if (x==y)                                                   \
    gp_substitution_matrix[x][y] =SUBMATRIXTABLE[matrix][0];   \
   else if ( (x%2) == (y%2))                                   \
      gp_substitution_matrix[x][y] =SUBMATRIXTABLE[matrix][2]; \
     else                                                      \
      gp_substitution_matrix[x][y] =SUBMATRIXTABLE[matrix][1]; \
   else                                                        \
      gp_substitution_matrix[x][y] =SUBMATRIXTABLE[matrix][3]; \
   gp_substitution_matrix[x][y]*=-1;                           \
   gp_cost_max_substitution_matrix =                           \
      MAX(                                                     \
	ABS(gp_substitution_matrix[x][y]),                     \
	gp_cost_max_substitution_matrix                        \
	);                                                     \
 }                                                             \
                                                               \
/* lambda/k*/                                                  \
if ( gp_k_blast == 0.0 && gp_lambda_blast == 0.0){             \
 gp_k_blast             = KBLASTTABLE[matrix];                 \
 gp_lambda_blast        = LAMBDABLASTTABLE[matrix];            \
}                                                              \


                                                               






/************************
 * DEBUT DES HOSTILITES *
 ************************/

MA * ma;

int main(int argc, char *argv[]){ 
  
  /* used by SCANARG() macro */
  int    i = 0;
  int    x;
  int    y;
  int    i_tmp = 0;
  double d_tmp = 0.0; 
  char   *s_tmp;
  int    matrix = 0;
  int overrides_weight = 0;
  FILE * f        = NULL ;
  /* first file chunks */
  int     nbchunks_query    = 0;
  char ** chunkname_query   = NULL;
  int  *  chunksize_query   = NULL;
  int  *  chunkstrt_query   = NULL;
  /* second file chunks */  
  int     nbchunks_text     = 0;
  char ** chunkname_text    = NULL;
  int  *  chunksize_text    = NULL;
  int  *  chunkstrt_text    = NULL;
  /* chunk */
  char *  query_chunk       = NULL;
  char *  queryrev_chunk    = NULL;
  int     querysize_chunk   = 0;
  
  /* [1] command line read */ 
  SCANARG();
  AutomataOverlaps(AUTOMATADELPH);
 
  /* [2] initiate parameters and statistics  */
  gp_mutations = (double) gp_mutations_percent / 100.000;
     gp_indels = (double)    gp_indels_percent / 100.000;  
      gp_alpha = (double)     gp_alpha_percent / 100.000;    
   gp_rho_stat = statistical_bound_of_waiting_time2((1-gp_mutations),gp_seed_bitweight/2,gp_alpha);	
 gp_delta_stat = statistical_bound_of_randomwalk2(gp_indels,gp_rho_stat,gp_alpha);  

 
 /* [3] chaining algorithm memory required */
 initialise_alignment(8*MEGA);
 initialise_deltashift();

 /* [4] read files and set their attributes */
 if (gp_nbfiles == 2 ){ 
   CreateData(gp_files[0],&gp_query,&gp_querysize,&nbchunks_query,&chunkname_query,&chunksize_query,&chunkstrt_query );
   CreateData(gp_files[1],&gp_text ,&gp_textsize, &nbchunks_text, &chunkname_text, &chunksize_text, &chunkstrt_text  );   
 }else{
   CreateData(gp_files[0],&gp_text ,&gp_textsize,&nbchunks_text, &chunkname_text, &chunksize_text ,&chunkstrt_text);
   /*->*/
   gp_query        = gp_text;
   gp_querysize    = gp_textsize;
   nbchunks_query  = nbchunks_text;
   chunkname_query = chunkname_text;
   chunksize_query = chunksize_text;
   chunkstrt_query = chunkstrt_text;
   /*<-*/
 }

 gp_minscore = MinScore(gp_k_blast, gp_lambda_blast, chunksize_query[gp_selection_fasta-1], chunkstrt_text[nbchunks_text-1]+chunksize_text[nbchunks_text-1], gp_expectation_value);

 if ( nbchunks_query < gp_selection_fasta){
   ERROR("selected fastachunk does not exist in \"",gp_files[0],"\" file , please check your -S parameter");
 }
 query_chunk        =  gp_query + chunkstrt_query[gp_selection_fasta-1];
 querysize_chunk    =  chunksize_query[gp_selection_fasta-1];
 gp_querysize_chunk =  querysize_chunk;


#ifdef STATS	
 time_spent = time(NULL);
#endif

 /* 
  * [5] Assemble algorithms 
  */ 

 /* [5.1] Forward Chaining */
 if (gp_reverse != 1){

   if (gp_nbfiles == 2 ){  
     MultiAssemble_Double( query_chunk, querysize_chunk, 
			   gp_text , gp_textsize , nbchunks_text, chunkname_text, chunksize_text, chunkstrt_text);  
   }else{
     if (nbchunks_query > 1){
       /* FIXME : need  MultiAssemble_Single which do no report comparison on the same diagonal neither the half top matrix */
       MultiAssemble_Double( query_chunk,  querysize_chunk,
			     gp_text , gp_textsize   , nbchunks_text, chunkname_text, chunksize_text, chunkstrt_text);
     }else{
       Assemble_Single(query_chunk,querysize_chunk);
     }
   }
 }
 
  /* [5.2] Complementairy Reverse Chaining if requested */
  if (gp_reverse){
    gv_reverse_repeat = 1;

    /* DataRev */
    CreateReverseComplement(query_chunk,querysize_chunk,&queryrev_chunk);
    
    if (gp_nbfiles == 2 ){
      
      /* Assemble */  
      MultiAssemble_Double(queryrev_chunk,querysize_chunk, 
			   gp_text, gp_textsize , nbchunks_text, chunkname_text, chunksize_text, chunkstrt_text);
    }else{
      /* Assemble */ 
      if (nbchunks_query > 1){
	/* FIXME : need  MultiAssemble_SingleRev which do no report comparison on the same diagonal neither the half top matrix */
	MultiAssemble_Double(queryrev_chunk,querysize_chunk,
			     gp_text, gp_textsize , nbchunks_text, chunkname_text, chunksize_text, chunkstrt_text);
      }else{
	Assemble_SingleRev(queryrev_chunk,query_chunk,querysize_chunk);
      }
    }
  }
  
  /*
   * [6] PostProcess MA 
   */
  
  gv_left_correction = 0;
      
  fprintf(stderr,"\n");
  fprintf(stderr,DRULE);
  fprintf(stderr,"\n searching   ...  finished\n");
  
  /* [6.1] Prunning low scoring MA */ 

  /*
   * fprintf(stderr," pruning   ..... ");  
   * if (gv_first_MA)
   *    Prune_MAList(&gv_first_MA,gp_expectation_value);
   *
   * fprintf(stderr," finished \n");
   */

  /* [6.2] Entropy filtering on MA */ 
  fprintf(stderr," filtering   ... ");
  if (gv_first_MA)
    EntropyFilter_MAList(&gv_first_MA,
			 query_chunk,queryrev_chunk,
			 gp_text,chunkstrt_text);
  fprintf(stderr," finished\n");
  
  /* [6.3] Regrouping MAs */
  fprintf(stderr," grouping   .... ");
  if (gv_first_MA) 
    main_multi_regroup (gv_first_MA);
  fprintf(stderr," finished\n");
  
  /* [6.4] Sorting results */
  fprintf(stderr," sorting   ..... ");
  if (gv_first_MA)
    ListSort_MAList(&gv_first_MA,&ma,0x40000000);
  fprintf(stderr," finished\n");

  
#ifdef STATS
  clock_post = clock() - last_clock;          	
  time_spent = time(NULL) - time_spent;
#endif


  /* [7] Display Results */
  Display_Alignements(querysize_chunk,
		      query_chunk,queryrev_chunk,
		      chunkname_query[gp_selection_fasta-1],		      
		      gp_text , nbchunks_text, chunkname_text, chunksize_text, chunkstrt_text);
  Display_Params();
  Display_Stats();
  return 0;  
}
